@@ -73,7 +73,7 @@ extern struct LayFuncs WinLf;
 
 extern int  use_hardstatus;
 extern int  MsgWait, MsgMinWait;
-extern int  Z0width, Z1width;
+extern const int  Z0width, Z1width;
 extern unsigned char *blank, *null;
 extern struct mline mline_blank, mline_null, mline_old;
 extern struct mchar mchar_null, mchar_blank, mchar_so;
@@ -1391,7 +1391,7 @@ int xs, ys, xe, ye, n, bce;
 {
   int i;
   int up;
-  int oldtop, oldbot;
+  int oldbot;
   int alok, dlok, aldlfaster;
   int missy = 0;
 
@@ -1430,7 +1430,6 @@ int xs, ys, xe, ye, n, bce;
   if (n >= ye - ys + 1)
     n = ye - ys + 1;
 
-  oldtop = D_top;
   oldbot = D_bot;
   if (ys < D_top || D_bot != ye)
     ChangeScrollRegion(ys, ye);
@@ -3446,7 +3445,7 @@ char *data;
   if (D_mouse && D_forecv)
     {
       unsigned char *bp = (unsigned char *)buf;
-      int x, y, i = size;
+      int x, y, i;
 
       /* XXX this assumes that the string is read in as a whole... */
       for (i = size; i > 0; i--, bp++)
