@@ -1,11 +1,16 @@
-/* Copyright (c) 1993-2002
+/* Copyright (c) 2008, 2009
+ *      Juergen Weigert (jnweiger@immd4.informatik.uni-erlangen.de)
+ *      Michael Schroeder (mlschroe@immd4.informatik.uni-erlangen.de)
+ *      Micah Cowan (micah@cowan.name)
+ *      Sadrul Habib Chowdhury (sadrul@users.sourceforge.net)
+ * Copyright (c) 1993-2002, 2003, 2005, 2006, 2007
  *      Juergen Weigert (jnweiger@immd4.informatik.uni-erlangen.de)
  *      Michael Schroeder (mlschroe@immd4.informatik.uni-erlangen.de)
  * Copyright (c) 1987 Oliver Laumann
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2, or (at your option)
+ * the Free Software Foundation; either version 3, or (at your option)
  * any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -14,16 +19,13 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program (see the file COPYING); if not, write to the
- * Free Software Foundation, Inc.,
- * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
+ * along with this program (see the file COPYING); if not, see
+ * http://www.gnu.org/licenses/, or contact Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA  02111-1301  USA
  *
  ****************************************************************
  */
 
-#include "rcs.h"
-RCS_ID("$Id: input.c,v 1.2 1994/05/31 12:32:08 mlschroe Exp $ FAU")
-
 #include <sys/types.h>
 #include "config.h"
 #include "screen.h"
@@ -41,23 +43,29 @@ extern struct mchar mchar_blank, mchar_so;
 
 struct inpline
 {
-  char  buf[101];	/* text buffer */
+  char  buf[MAXSTR+1];	/* text buffer */
   int  len;		/* length of the editible string */
   int  pos;		/* cursor position in editable string */
+  struct inpline *next, *prev;
 };
 
-static struct inpline inphist; /* XXX: should be a dynamic list */
-
+/* 'inphist' is used to store the current input when scrolling through history.
+ * inpline->prev == history-prev
+ * inpline->next == history-next
+ */
+static struct inpline inphist;
 
 struct inpdata
 {
   struct inpline inp;
-  int  inpmaxlen;	/* 100, or less, if caller has shorter buffer */
+  int  inpmaxlen;	/* MAXSTR, or less, if caller has shorter buffer */
   char *inpstring;	/* the prompt */
   int  inpstringlen;	/* length of the prompt */
   int  inpmode;		/* INP_NOECHO, INP_RAW, INP_EVERY */
   void (*inpfinfunc) __P((char *buf, int len, char *priv));
   char  *priv;		/* private data for finfunc */
+  int  privdata;	/* private data space */
+  char *search; 	/* the search string */
 };
 
 static struct LayFuncs InpLf =
@@ -68,7 +76,8 @@ static struct LayFuncs InpLf =
   DefClearLine,
   DefRewrite,
   DefResize,
-  DefRestore
+  DefRestore,
+  0
 };
 
 /*
@@ -96,6 +105,8 @@ char *p, *s;
       inpdata->inp.pos = inpdata->inp.len = strlen(inpdata->inp.buf);
     }
   InpRedisplayLine(INPUTLINE, 0, flayer->l_width - 1, 0);
+  flayer->l_x = inpdata->inpstringlen + (inpdata->inpmode & INP_NOECHO ? 0 : inpdata->inp.pos);
+  flayer->l_y = INPUTLINE;
 }
 
 /*
@@ -109,18 +120,22 @@ char *p, *s;
  * INP_EVERY  == digraph mode.
  */
 void
-Input(istr, len, mode, finfunc, data)
+Input(istr, len, mode, finfunc, priv, data)
 char *istr;
 int len;
 int mode;
-void (*finfunc) __P((char *buf, int len, char *data));
-char *data;
+void (*finfunc) __P((char *buf, int len, char *priv));
+char *priv;
+int data;
 {
   int maxlen;
   struct inpdata *inpdata;
   
-  if (len > 100)
-    len = 100;
+  if (!flayer)
+    return;
+
+  if (len > MAXSTR)
+    len = MAXSTR;
   if (!(mode & INP_NOECHO))
     {
       maxlen = flayer->l_width - 1 - strlen(istr);
@@ -134,15 +149,58 @@ char *data;
     }
   if (InitOverlayPage(sizeof(*inpdata), &InpLf, 1))
     return;
+  flayer->l_mode = 1;
   inpdata = (struct inpdata *)flayer->l_data;
   inpdata->inpmaxlen = len;
   inpdata->inpfinfunc = finfunc;
   inpdata->inp.pos = inpdata->inp.len = 0;
+  inpdata->inp.prev = inphist.prev;
   inpdata->inpmode = mode;
-  inpdata->priv = data;
-  inp_setprompt(istr, (char *)NULL);
-  flayer->l_x = inpdata->inpstringlen;
-  flayer->l_y = INPUTLINE;
+  inpdata->privdata = data;
+  if (!priv)
+    priv = (char*)&inpdata->privdata;
+  inpdata->priv = priv;
+  inpdata->inpstringlen = 0;
+  inpdata->inpstring = NULL;
+  inpdata->search = NULL;
+  if (istr)
+    inp_setprompt(istr, (char *)NULL);
+}
+
+static void
+erase_chars(inpdata, from, to, x, mv)
+struct inpdata *inpdata;
+char *from;
+char *to;
+int x;
+int mv;
+{
+  int chng;
+  ASSERT(from < to);
+  if (inpdata->inp.len > to - inpdata->inp.buf)
+    bcopy(to, from, inpdata->inp.len - (to - inpdata->inp.buf));
+  chng = to - from;
+  if (mv)
+    {
+      x -= chng;
+      inpdata->inp.pos -= chng;
+    }
+  inpdata->inp.len -= chng;
+  if (!(inpdata->inpmode & INP_NOECHO))
+    {
+      struct mchar mc;
+      char *s = from < to ? from : to;
+      mc = mchar_so;
+      while (s < inpdata->inp.buf+inpdata->inp.len)
+	{
+	  mc.image = *s++;
+	  LPutChar(flayer, &mc, x++, INPUTLINE);
+	}
+      while (chng--)
+	LPutChar(flayer, &mchar_blank, x++, INPUTLINE);
+      x = inpdata->inpstringlen + inpdata->inp.pos;
+      LGotoPos(flayer, x, INPUTLINE);
+    }
 }
 
 static void
@@ -155,10 +213,13 @@ int *plen;
   char ch;
   struct inpdata *inpdata;
   struct display *inpdisplay;
+  int prev, next, search = 0;
 
   inpdata = (struct inpdata *)flayer->l_data;
   inpdisplay = display;
 
+#define RESET_SEARCH do { if (inpdata->search) Free(inpdata->search); } while (0)
+
   LGotoPos(flayer, inpdata->inpstringlen + (inpdata->inpmode & INP_NOECHO ? 0 : inpdata->inp.pos), INPUTLINE);
   if (ppbuf == 0)
     {
@@ -177,10 +238,12 @@ int *plen;
       if (inpdata->inpmode & INP_EVERY)
 	{
 	  inpdata->inp.buf[inpdata->inp.len] = ch;
-	  inpdata->inp.buf[inpdata->inp.len + 1] = ch;	/* gross */
-	  display = inpdisplay;
-	  (*inpdata->inpfinfunc)(inpdata->inp.buf, inpdata->inp.len, inpdata->priv);
-	  ch = inpdata->inp.buf[inpdata->inp.len];
+	  if (ch)
+	    {
+	      display = inpdisplay;
+	      (*inpdata->inpfinfunc)(inpdata->inp.buf, inpdata->inp.len, inpdata->priv);
+	      ch = inpdata->inp.buf[inpdata->inp.len];
+	    }
 	}
       else if (inpdata->inpmode & INP_RAW)
 	{
@@ -214,33 +277,16 @@ int *plen;
 		  LGotoPos(flayer, x, INPUTLINE);
 		}
 	    }
+	  RESET_SEARCH;
 	}
       else if ((ch == '\b' || ch == 0177) && inpdata->inp.pos > 0)
 	{
-	  if (inpdata->inp.len > inpdata->inp.pos)
-	    bcopy(p, p-1, inpdata->inp.len - inpdata->inp.pos);
-	  inpdata->inp.len--;
-	  inpdata->inp.pos--;
-	  p--;
-
-	  if (!(inpdata->inpmode & INP_NOECHO))
-	    {
-	      struct mchar mc;
-	      mc = mchar_so;
-	      x--;
-	      while (p < inpdata->inp.buf+inpdata->inp.len)
-		{
-		  mc.image = *p++;
-		  LPutChar(flayer, &mc, x++, INPUTLINE);
-		}
-	      LPutChar(flayer, &mchar_blank, x, INPUTLINE);
-	      x = inpdata->inpstringlen + inpdata->inp.pos;
-	      LGotoPos(flayer, x, INPUTLINE);
-	    }
+	  erase_chars(inpdata, p-1, p, x, 1);
+	  RESET_SEARCH;
 	}
       else if (ch == '\025')			/* CTRL-U */
 	{
-  	  x = inpdata->inpstringlen;
+	  x = inpdata->inpstringlen;
 	  if (inpdata->inp.len && !(inpdata->inpmode & INP_NOECHO))
 	    {
 	      LClearArea(flayer, x, INPUTLINE, x + inpdata->inp.len - 1, INPUTLINE, 0, 0);
@@ -250,7 +296,7 @@ int *plen;
 	}
       else if (ch == '\013')			/* CTRL-K */
 	{
-  	  x = inpdata->inpstringlen + inpdata->inp.pos;
+	  x = inpdata->inpstringlen + inpdata->inp.pos;
 	  if (inpdata->inp.len > inpdata->inp.pos && !(inpdata->inpmode & INP_NOECHO))
 	    {
 	      LClearArea(flayer, x, INPUTLINE, x + inpdata->inp.len - inpdata->inp.pos - 1, INPUTLINE, 0, 0);
@@ -258,6 +304,21 @@ int *plen;
 	    }
 	  inpdata->inp.len = inpdata->inp.pos;
 	}
+      else if (ch == '\027' && inpdata->inp.pos > 0)		/* CTRL-W */
+	{
+	  char *oldp = p--;
+	  while (p > inpdata->inp.buf && *p == ' ')
+	    p--;
+	  while (p > inpdata->inp.buf && *(p - 1) != ' ')
+	    p--;
+	  erase_chars(inpdata, p, oldp, x, 1);
+	  RESET_SEARCH;
+	}
+      else if (ch == '\004' && inpdata->inp.pos < inpdata->inp.len)	/* CTRL-D */
+	{
+	  erase_chars(inpdata, p, p+1, x, 0);
+	  RESET_SEARCH;
+	}
       else if (ch == '\001' || (unsigned char)ch == 0201)	/* CTRL-A */
 	{
 	  LGotoPos(flayer, x -= inpdata->inp.pos, INPUTLINE);
@@ -278,22 +339,57 @@ int *plen;
 	  LGotoPos(flayer, ++x, INPUTLINE);
 	  inpdata->inp.pos++;
 	}
-      else if (ch == '\020' || (unsigned char)ch == 0220)	/* CTRL-P */
-        {
+      else if ((prev = ((ch == '\020' || (unsigned char)ch == 0220) &&	/* CTRL-P */
+	      inpdata->inp.prev)) ||
+	  (next = ((ch == '\016' || (unsigned char)ch == 0216) &&  /* CTRL-N */
+		   inpdata->inp.next)) ||
+	  (search = ((ch == '\022' || (unsigned char)ch == 0222) && inpdata->inp.prev)))
+	{
 	  struct mchar mc;
+	  struct inpline *sel;
+	  int pos = -1;
+
 	  mc = mchar_so;
+
+	  if (prev)
+	    sel = inpdata->inp.prev;
+	  else if (next)
+	    sel = inpdata->inp.next;
+	  else
+	    {
+	      /* search */
+	      inpdata->inp.buf[inpdata->inp.len] = 0;	/* Remove the ctrl-r from the end */
+	      if (!inpdata->search)
+		inpdata->search = SaveStr(inpdata->inp.buf);
+	      for (sel = inpdata->inp.prev; sel; sel = sel->prev)
+		{
+		  char *f;
+		  if ((f = strstr(sel->buf, inpdata->search)))
+		    {
+		      pos = f - sel->buf;
+		      break;
+		    }
+		}
+	      if (!sel)
+		continue;	/* Did not find a match. Process the next input. */
+	    }
+
 	  if (inpdata->inp.len && !(inpdata->inpmode & INP_NOECHO))
 	    LClearArea(flayer, inpdata->inpstringlen, INPUTLINE, inpdata->inpstringlen + inpdata->inp.len - 1, INPUTLINE, 0, 0);
 
-	  inpdata->inp = inphist;	/* structure copy */
+	  if ((prev || search) && !inpdata->inp.next)
+	    inphist = inpdata->inp;
+	  memcpy(&inpdata->inp, sel, sizeof(struct inpline));
+	  if (pos != -1)
+	    inpdata->inp.pos = pos;
 	  if (inpdata->inp.len > inpdata->inpmaxlen)
 	    inpdata->inp.len = inpdata->inpmaxlen;
 	  if (inpdata->inp.pos > inpdata->inp.len)
 	    inpdata->inp.pos = inpdata->inp.len;
 
-  	  x = inpdata->inpstringlen;
+	  x = inpdata->inpstringlen;
 	  p = inpdata->inp.buf;
-	  
+
 	  if (!(inpdata->inpmode & INP_NOECHO))
 	    {
 	      while (p < inpdata->inp.buf+inpdata->inp.len)
@@ -302,21 +398,48 @@ int *plen;
 		  LPutChar(flayer, &mc, x++, INPUTLINE);
 		}
 	    }
-  	  x = inpdata->inpstringlen + inpdata->inp.pos;
+	  x = inpdata->inpstringlen + inpdata->inp.pos;
 	  LGotoPos(flayer, x, INPUTLINE);
 	}
 
-      else if (ch == '\004' || ch == '\003' || ch == '\007' || ch == '\033' ||
+      else if (ch == '\003' || ch == '\007' || ch == '\033' ||
 	       ch == '\000' || ch == '\n' || ch == '\r')
 	{
-          if (ch != '\004' && ch != '\n' && ch != '\r')
+          if (ch != '\n' && ch != '\r')
 	    inpdata->inp.len = 0;
 	  inpdata->inp.buf[inpdata->inp.len] = 0;
 
-	  if (inpdata->inp.len && inpdata->inpmode == 0)
-	    inphist = inpdata->inp;	/* structure copy */
-	  
-  	  flayer->l_data = 0;
+	  if (inpdata->inp.len && !(inpdata->inpmode & (INP_NOECHO | INP_RAW)))
+	    {
+	      struct inpline *store;
+
+	      /* Look for a duplicate first */
+	      for (store = inphist.prev; store; store = store->prev)
+		{
+		  if (strcmp(store->buf, inpdata->inp.buf) == 0)
+		    {
+		      if (store->next)
+			store->next->prev = store->prev;
+		      if (store->prev)
+			store->prev->next = store->next;
+		      store->pos = inpdata->inp.pos;
+		      break;
+		    }
+		}
+
+	      if (!store)
+		{
+		  store = malloc(sizeof(struct inpline));
+		  memcpy(store, &inpdata->inp, sizeof(struct inpline));
+		}
+	      store->next = &inphist;
+	      store->prev = inphist.prev;
+	      if (inphist.prev)
+		inphist.prev->next = store;
+	      inphist.prev = store;
+	    }
+
+	  flayer->l_data = 0;	/* so inpdata does not get freed */
           InpAbort();		/* redisplays... */
 	  *ppbuf = pbuf;
 	  *plen = len;
@@ -325,9 +448,17 @@ int *plen;
             (*inpdata->inpfinfunc)(inpdata->inp.buf, inpdata->inp.len, inpdata->priv);
 	  else
             (*inpdata->inpfinfunc)(pbuf - 1, 0, inpdata->priv);
-	  free((char *)inpdata);
+	  if (inpdata->search)
+	    free(inpdata->search);
+	  free(inpdata);
 	  return;
 	}
+      else
+	{
+	  /* The user was searching, and then pressed some non-control input. So reset
+	   * the search string. */
+	  RESET_SEARCH;
+	}
     }
   if (!(inpdata->inpmode & INP_RAW))
     {
@@ -351,7 +482,7 @@ int y, xs, xe, isblank;
 {
   int q, r, s, l, v;
   struct inpdata *inpdata;
-  
+
   inpdata = (struct inpdata *)flayer->l_data;
   if (y != INPUTLINE)
     {
@@ -395,11 +526,3 @@ int y, xs, xe, isblank;
     }
 }
 
-int
-InInput()
-{
-  if (flayer && flayer->l_layfn == &InpLf)
-    return 1;
-  return 0;
-}
-
