@@ -1019,6 +1019,8 @@ int wi, he, hi;
   /* store new size */
   p->w_width = wi;
   p->w_height = he;
+  if(p->w_scrollback_height > hi)
+    p->w_scrollback_height = hi;
 #ifdef COPY_PASTE
   p->w_histidx = 0;
   p->w_histheight = hi;
@@ -1091,6 +1093,14 @@ struct win *p;
   SWAP(width, t);
   SWAP(height, t);
 
+  /* we need to swap cursor positions, but they are not swappable with macro */
+  t = p->w_alt.cursor.x;
+  p->w_alt.cursor.x = p->w_x;
+  p->w_x = t;
+  t = p->w_alt.cursor.y;
+  p->w_alt.cursor.y = p->w_y;
+  p->w_y = t;
+
 #ifdef COPY_PASTE
   SWAP(histheight, t);
   SWAP(hlines, ml);
